package nxm.sys.intr;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import java.util.regex.Pattern;

import nxm.sys.inc.Chainable;
import nxm.sys.inc.Keyable;
import nxm.sys.lib.Args;
import nxm.sys.lib.Command;
import nxm.sys.lib.Intrinsic;
import nxm.sys.lib.KeyObject;
import nxm.sys.lib.Parser;
import nxm.sys.lib.Shell;
import nxm.sys.lib.StringUtil;
import nxm.sys.lib.Terminal;

/**
  Queries an Object for public constructors, methods, fields, and interfaces.

  @author Jeff Schoen/Jay Hallam
*/
public class Query extends Intrinsic {

  /** filter list for constructors, methods, and fields */
  public static final String filterList = "Public,Private,Inherited,Interface,FullSig,Docs";
  private static int F_NONE=0, F_PUBLIC=0x1, F_PRIVATE=0x2, F_INHERIT=0x4, F_INTER=0x8;
  private static int F_FULLSIG=0x10, F_DOCS=0x20;
  private static int F_ALL=0xF;  // not used, delete?

  private int constMask,fieldMask,methodMask;

  private int numRows;
  private int colWidth;
  private String memberName;
  private String listName,listValue;

  @SuppressWarnings("deprecation")  // for M.terminal - since 3.7.1 for NXM-2813
  public int process () {
    listValue=null;

    Class<?> clas=null;
    Object object=null;

    // JDK 7+ uses CSS for Javadocs, so Query cannot properly render them (since 3.3.1 - Bug 2653)
    boolean jdk7plus = (StringUtil.compareVersions("1.7.0", Shell.getJavaVersion()) <= 0);

    // Get parameters
    String csName  = MA.getCS("NAME");  // If case-sensitive value is needed
    String resName = MA.getU("NAME"); 	// If needed as a result name
    listName       = MA.getU("/LIST",null); 	// If needed as a result name
    memberName     = MA.getCS("METH");
    boolean all    = MA.getState("/ALL");

     // limited functionality in jdk7+ (since 3.3.1 - Bug 2653)
    if (!jdk7plus) {
      int defMask = F_PUBLIC | F_INHERIT;
      if (all) defMask |= (F_PRIVATE | F_INTER);
      defMask = MA.getOptionMask("/MASK",filterList,defMask);

      boolean doMatch = MA.isPresent("METH") && !memberName.equals("*");
      if (doMatch) { defMask |= (F_FULLSIG|F_DOCS); }
      fieldMask = constMask = methodMask = defMask;

      if (MA.getState("/GETS")) { memberName = "GET*"; constMask = fieldMask = F_NONE; }
      if (MA.getState("/SETS")) { memberName = "SET*"; constMask = fieldMask = F_NONE; }
      if (MA.getState("/FIELDS")) { methodMask = constMask = F_NONE; }

      methodMask = MA.getOptionMask("/M",filterList,methodMask);
      fieldMask  = MA.getOptionMask("/F",filterList,fieldMask);
      constMask  = MA.getOptionMask("/C",filterList,constMask);
    } else {
      fieldMask = constMask = methodMask = F_ALL;
    }

    if (MA.getState("/DEBUG")) {
      System.out.println("Query Name="+csName+" matche="+memberName+" MethodMask="+methodMask
                        +" FieldMask="+fieldMask+" all="+all);
    }

    // get the object
    if (csName.length()==0) {					// no arg
      M.error("No object specified");
    } else if ((object=MR.get(resName))!=null) {		// try result
      clas = object.getClass();
    } else {
      try {							// try class name
        clas = Class.forName(csName);
      } catch (Exception e) {					// try command arg
        Args targs = Shell.parseCommand(M,csName);
        Command command = Shell.loadCommand(M,targs);
        switch(command.support) {
          case 'P': clas = command.getClass(); object = command; break;
          case 'I': clas = command.getClass(); object = command; break;
          default:  clas = getClass(csName); break;
        }
      }
    }

    // setup terminal output more filter
    colWidth = MA.getL("/NC",colWidth);             // Number of cols in more filter
    numRows  = MA.getL("/NR",M.terminal.getRows()); // Number of rows in more filter
    if (MA.find("/NR")) M.terminal.setRows(numRows);// Number of rows in more filter

    if (clas==null) {
      M.error("Query Object "+csName+" not found");
    } else {
        String fileName  = getJavaDocsFilename(clas.getName().trim());
        if (fileName == null) fileName = "API location.";
        M.info("For best viewing, view API in web browser at "+fileName);
      if (jdk7plus) {  // for JDK 7+, let the user know the functionality is limited and point them to API web page
        M.warning("Use of QUERY with JDK 7+ has limited functionality.");
      }
    }

    // perform the query
    if (listName!=null) {
      doQueryClass (clas, memberName);
      MR.put(listName,listValue);
    }
    else {
      doQuery(clas, object);
    }

    return FINISH;
  }

  /** Get the API filename for the class being queried
      @param fileName String form of the class name
      @return String with path to API (e.g. /nmroot/htdocs/api/sys/nxm/sys/lib/Table.html
      @since NeXtMidas 3.3.1 (added for Bug 2653)
   */
  private String getJavaDocsFilename (String fname) {
    // this code moved from getJavadocs() so it can be used in process() also
    if (fname==null) return null;
    String fileName = fname;
    if (fileName.startsWith("nxm.")) {
      final String opt = fileName.substring(4, fileName.indexOf('.', 4));
      fileName = Shell.getNmRoot()+
          "htdocs" + java.io.File.separator +
          "api"    + java.io.File.separator +
          opt      + java.io.File.separator +
          fileName;
      fileName  = fileName.replace('.', java.io.File.separatorChar) + ".html";
    } else {
      fileName = null;
    }
    return fileName;
  }

  /** Gets the next Object down the chain from the given Object
      @param object the Object to be analyzed
      @return the next Object down the chain from the given Object
   */
  public Object nextChainedObject (Object object) {
    Object nextObject = null;
    if (object instanceof Chainable) {
      nextObject = ((Chainable)object).getNextLink();
    }
    return nextObject;
  }

  /** Handles the query.
      @param clas   The class.
      @param object The object (used when finding chainable methods).
   */
  @SuppressWarnings("deprecation")  // for M.terminal - since 3.7.1 for NXM-2813
  public void doQuery (Class<?> clas, Object object) {
    M.terminal.more(Terminal.INIT);
    doQuery(clas, object, "ClassName");
    M.terminal.more(Terminal.EXIT);
  }

  /** Handles the query. Used only by doQuery(..).
      @param clas   The class.
      @param object The object.
      @param title  The title of the class ("ClassName" or "Chained to Class").
   */
  @SuppressWarnings("deprecation")  // for M.terminal - since 3.7.1 for NXM-2813
  private void doQuery (Class<?> clas, Object object, String title) {
    Object nextObj = nextChainedObject(object);

    M.more(title+" -> "+clas.getName());  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
    if (Keyable.class.isAssignableFrom(clas)) {
      M.more(" *** Class implements Keyable, setKey and getKey may override methods ***");
    }
    doQueryClass (clas, memberName);
    M.more(Terminal.FLUSH);

    if (nextObj != null) {
      M.more("");
      doQuery(nextObj.getClass(), nextObj, "Chained to Class");
    }
  }

  /** Lists all the constructors, methods and fields.
      @param clazz   The class.
      @param name    Name of the members to show ("*" for all).
   */
  @SuppressWarnings({"unchecked", "deprecation"})
  private void doQueryClass (Class<?> clazz, String name) {
    String constHeader  = "Constructors - all declared with options "+Parser.mask2s(filterList,constMask);
    String methodHeader = "Methods - all declared with options "+Parser.mask2s(filterList,methodMask);
    String fieldHeader  = "Fields - all declared with options "+Parser.mask2s(filterList,fieldMask);

    List<Constructor<?>> consts  = new ArrayList<>();
    List<Method> methods = new ArrayList<>();
    List<Field> fields = new ArrayList<>();

    if (name.indexOf("*")>=0) {

      /* Add the public ones */
      consts.addAll(Arrays.asList(clazz.getConstructors()));
      methods.addAll(Arrays.asList(clazz.getMethods()));
      fields.addAll(Arrays.asList(clazz.getFields()));

      /* Add the private, protected ones. */
      consts.addAll(  (List<Constructor<?>>) removePublic(clazz.getDeclaredConstructors()));
      methods.addAll( (List<Method>) removePublic(clazz.getDeclaredMethods()));
      fields.addAll(  (List<Field>) removePublic(clazz.getDeclaredFields()));

      /* Sort */
      String[] sortedConsts = sortNames(clazz,consts.toArray(new Member[0]),  name, constMask);
      String[] sortedMethods= sortNames(clazz,methods.toArray(new Member[0]), name, methodMask);
      String[] sortedFields = sortNames(clazz,fields.toArray(new Member[0]),  name, fieldMask);

      /* List everything */
      listInfo(constHeader,  sortedConsts,  clazz, constMask);
      listInfo(methodHeader, sortedMethods, clazz, methodMask);
      listInfo(fieldHeader,  sortedFields,  clazz, fieldMask);
    }
    else if (clazz.equals(nxm.sys.lib.Table.class) || clazz.equals(nxm.sys.lib.KeyVector.class)) {
      M.more(" *** Only getKey and setKey can be used ***");  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
      M.more(Terminal.FLUSH);
    }
    else if (clazz.isArray()) {
      M.more(" *** Only getSize() and array access methods can be used ***");  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
      M.more(Terminal.FLUSH);
    }
    else {
      String     key     = (name.equals("*"))? null: name;
      Class<?>[] types   = null;
      Member[]   members = KeyObject.findMembers(clazz, key, types, false);

      for (Member member : members) {
        if (member instanceof Constructor) consts.add((Constructor<?>)member);
        else if (member instanceof Method) methods.add((Method)member);
        else if (member instanceof Field)  fields.add((Field)member);
      }

      /* Sort */
      String[] sortedConsts = sortNames(clazz,consts.toArray(new Member[0]),  name, constMask);
      String[] sortedMethods= sortNames(clazz,methods.toArray(new Member[0]), name, methodMask);
      String[] sortedFields = sortNames(clazz,fields.toArray(new Member[0]),  name, fieldMask);

      /* List everything */
      listInfo(constHeader,  sortedConsts,  clazz, constMask);
      listInfo(methodHeader, sortedMethods, clazz, methodMask);
      listInfo(fieldHeader,  sortedFields,  clazz, fieldMask);
    }

  }

  /** Filters out all public members.
      @param members The items.
      @return A list of the non-public members.
   */
  private static List<? extends Member> removePublic (Member[] members) {
    List<Member> list = new ArrayList<>();

    for (Member member: members) {
      if (!Modifier.isPublic(member.getModifiers())) {
        list.add(member);
      }
    }

    return list;
  }


  /** Lists the information about a set of constructors/methods/fields.
      @param header The header.
      @param list   The list of the ones to list.
      @param clazz  The relevant class.
   */
  @SuppressWarnings("deprecation")  // for M.terminal - since 3.7.1 for NXM-2813
  private void listInfo (String header, String[] list, Class<?> clazz, int mask) {

    if (mask==0) return;

    if (listName!=null) {
      for (int i=0; i<list.length; i++) {
        if (listValue==null) listValue=list[i];
        else listValue = listValue+","+list[i];
      }
      return;
    }

    M.more(" " + header + " (" + list.length + ") -> ");  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813

    for (int i=0; i<list.length; i++) {
      if ((mask&F_DOCS)!=0) {                                     /* List the javadocs */
        if (i != 0) { M.more("----------"); }
        String[] javadocs = getJavadocs(list[i], clazz.getName());
        listDocs (javadocs);
      } else {                                                    /* List the name */
        M.more("  "+list[i],colWidth);  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
      }
    }
    M.more(Terminal.FLUSH);  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
  }

  @SuppressWarnings("deprecation")  // for M.terminal - since 3.7.1 for NXM-2813
  private void listDocs (String[] javadocs) {
    for (int line = 0; line < javadocs.length; line++) { // go through docs line by line

      String text = javadocs[line];
      int    max  = Math.min(100, M.terminal.getColumns()) - 1;

      for (int xx=0; xx < text.length();) {  // allow for lines greater than the col width
        if (text.length() < max) {
          M.more(text);  // since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
          text = "";
          xx = text.length();  // for final printed line, get out of FOR loop
        } else {
          int index = text.lastIndexOf(' ', max);  // find last white space within max chars
          String indent = "      ";   // number of spaces to indent
          if ((xx < 1) && (index < 1)) { index = max; } //first time through
          else if ((xx >= 1) && (index < indent.length())) index = max; //after first time, allow for indent
          M.more(text.substring(0, index));   // print the line - since 3.7.1 using M.more instead of M.terminal.more - NXM-1813
          text = indent + text.substring(index).trim();  // remove the printed line from the remainder
          xx++;  // increment counter
        }
      }
    }
  }



  /** Gets the javadocs for a given method. <br>
      This method makes heavy use of regular expressions to parse the HTML file
      generated by the standard doclet used by javadoc. This will only work with
      the standard doclet. The processed text is filtered to remove HTML tags
      but such is just a basic filter and some of the formatting may be lost.
      @param methodSig The method signature (e.g. <code>"public static void
                       main(java.lang.String[])"</code> or <code>"public static
                       void foo.bar.Baz.main(java.lang.String[])"</code>).
      @param clazz     The fully-resolved name of the class the method belongs
                       to (e.g. <code>"java.lang.String"</code>).
      @return The lines that make up the javadoc comment for the method.
   */
  private String[] getJavadocs (String methodSig, String clazz) {
    methodSig = methodSig.trim();
    clazz     = clazz.trim();

    final boolean isParam = (methodSig.indexOf('(') < 0);

    int    open    = methodSig.indexOf('(');
    int    close   = methodSig.indexOf(')');
    String params  = isParam? ""        : methodSig.substring(open+1,close);
    String prefix  = isParam? methodSig : methodSig.substring(0, open);
    int    start   = Math.max(0, prefix.lastIndexOf(' '));
    String name    = prefix.substring(start);
    int    dot     = Math.max(0, name.lastIndexOf('.'));
    String method  = name.substring(dot + 1);

    String paramsList = params.replaceAll(",",   ",\\\\s*")  // Allow whitespace after ","
                              .replaceAll("\\[", "\\\\[")    // Change "[" to "\["
                              .replaceAll("\\]", "\\\\]");   // Change "[" to "\["
    String regex = method + (isParam? "": "[(]" + paramsList + "[)]");

    if (dot > 0) {
      if (!clazz.equals(name.trim())) {
        clazz = name.substring(0, dot).trim();
      }
    }

    //final String fileName  = clazz.replace('.', java.io.File.separatorChar) + ".html";

    java.io.File file = null;

    if (clazz.startsWith("nxm.")) {
      file = new java.io.File(getJavaDocsFilename(clazz));
    }

    if ((file == null) || (!file.exists())) {
        return new String[] { "  " + methodSig + "\n",
                              "    Javadoc file not found" };

    } else {
      try {
        String javadoc = " ";

        java.io.BufferedReader input = new java.io.BufferedReader(new java.io.FileReader(file));
        Pattern top = Pattern.compile(".*<[Aa]\\s+[Nn][Aa][Mm][Ee]="
                                    + "((\"" + regex + "\")"
                                    + "|('"  + regex +  "')"
                                    + "|("   + regex +   "))"
                                    + "\\s*>.*");
        Pattern btm = Pattern.compile(".*<[Hh][Rr]>.*");

        String line = null;

        /* Go to the top of method comment. */
        while ((line = input.readLine()) != null) {
          if (top.matcher(line).matches()) {
            line = input.readLine();
            break;
          }
        }

        /* If file does not include javadocs for given method, just return the
           signature.
         */
        if (line == null) {
          input.close();
          return new String[] { "  " + methodSig + "\n","    No Javadoc found" };
        }

        /* Process until at the bottom of method comment. */
        while ((line = input.readLine()) != null) {
          if (btm.matcher(line).matches()) {
            break;
          }

          javadoc += line.replaceAll("<[Dd][Dd]>", "\n")
                         .replaceAll("<[Dd][Tt]>", "\n")
                         .replaceAll("<[Bb][Rr]>", "\n")
                         .replaceAll("<[^>]+>",    "")
                         .replaceAll("&nbsp;",     " ")
                         .replaceAll("&lt;",       "<")
                         .replaceAll("&gt;",       ">")
                         .replaceAll("&amp;",      "&")
                         .replaceAll("[ \t]+",     " ")
                  + " ";

        }
        input.close();

        javadoc.replaceAll("\n\\s*", "\n");

        final String[] jdLines = javadoc.split("\n");

        for (int i = 1; i < jdLines.length; i++) {
          jdLines[i] = jdLines[i].trim();

          if (jdLines[i].startsWith("Parameters:")
             || jdLines[i].startsWith("Returns:")
             || jdLines[i].startsWith("Overrides:")
             || jdLines[i].startsWith("Specified By:")
             || jdLines[i].startsWith("Deprecated:")
             || jdLines[i].startsWith("See Also:")
             || jdLines[i].startsWith("Throws:")
             || jdLines[i].startsWith("Author:")
             || jdLines[i].startsWith("Since:")
             || jdLines[i].startsWith("To Do:")) {
            jdLines[i] = "    " + jdLines[i];
          } else {
            jdLines[i] = "      " + jdLines[i];
          }
        }

        return jdLines;

      } catch(java.io.IOException e) {
        return new String[] { methodSig + "\n",
                              "  Javadoc file not available" };
      }
    }
  }


  /** Returns the names for a set of members.
      @param members The members.
      @param fullSig Indicates if the full signature should be listed.
      @return The names
   */
  @SuppressWarnings("unused")
  private static String[] getNames (Vector<Member> members, boolean fullSig) {
    String[] names = new String[members.size()];

    if (fullSig) {
      for (int i = 0; i < names.length; i++) {
        names[i] = members.get(i).toString();
      }
    } else {
      for (int i = 0; i < names.length; i++) {
        names[i] = members.get(i).getName();
      }
    }

    return names;
  }

  private static Class<?>[] getTypes (String args) {
    Class<?>[] params = null;

    args = args.trim();

    if (args.startsWith("(") && args.endsWith(")")) {
      args = args.substring(1,args.length()-1);
    }

    String[] paramStrs=args.split(",");

    if ((paramStrs.length == 1) && (paramStrs[0].trim().length() == 0)) {
      params = new Class<?>[0];

    } else {
      params = new Class<?>[paramStrs.length];

      for (int i = 0; i < params.length; i++) {
        params[i] = getClass(paramStrs[i].trim());

        if (params[i] == null) {
          Shell.warning("Unknown type '" + paramStrs[i] + "'.");
          params = null;
          break;
        }
      }
    }

    return params;
  }

  /** Load and sort an array of Methods, Fields or Classes into an ascending list of String names.
      @param objArray   The list of members to sort.
      @param memberName The name of the member to get.
      @param mask       The mask of filter options.
      @return The sorted list of members that match the given name.
   */
  private static String[] sortNames (Class<?> clazz, Member[] objArray, String memberName, int mask) {
    // Load an array of Strings with the names
    ArrayList<String> nameList = new ArrayList<>();

    final boolean all     = memberName.equals("*");         /* List all */
    final boolean exact   = (memberName.indexOf('(') >= 0); /* Exact match */
    final boolean fullSig = (mask&F_FULLSIG)!=0;

    int istar = memberName.indexOf("*");
    if (istar>0) memberName = memberName.substring(0,istar); /* match up to the "*" */

    /* Get the list of param names if provided. */
    Class<?>[] params = null;
    if (exact) {
      final int pStart = memberName.indexOf('(');
      final int pEnd   = memberName.indexOf(')');
      if (pEnd > 0) {
        params = getTypes(memberName.substring(pStart+1,pEnd));
      }
      memberName = memberName.substring(0, pStart).trim();
    }

    final String memberNameUC = memberName.toUpperCase();
    final String memberNameCS = memberName;

    try {
      for (int i = 0; i < objArray.length; i++) {
        Member member =  objArray[i];
        String sig  = member.toString();      // The member signature
        String name = member.getName();       // The member name
        int mods    = member.getModifiers();  // The member modifiers
        String signame = fullSig? sig : name;
        boolean isInherited = member.getDeclaringClass() != clazz;
        boolean isInterface = isInherited && member.getDeclaringClass().isInterface();
        if (isInterface) isInherited = false;

        if (member instanceof Constructor) name = "<init>";

             if (Modifier.isPrivate(mods) && (mask&F_PRIVATE)==0);
        else if (Modifier.isPublic(mods) && (mask&F_PUBLIC)==0);
        else if (isInherited && (mask&F_INHERIT)==0);
        else if (isInterface && (mask&F_INTER)==0);
        else if (all) {
          nameList.add(signame);
        }
        else if ((member instanceof Field) && name.equals(memberNameCS)) {
          nameList.add(signame);
        }
        else if (exact && name.equals(memberNameCS)) {
          if (params != null) {
            Class<?>[] methParams = null;
            if (member instanceof Method) {
              methParams = ((Method)member).getParameterTypes();
            }
            else if (member instanceof Constructor) {
              methParams = ((Constructor<?>)member).getParameterTypes();
            }
            else {
              nameList.add(sig);
              break;
            }
            if (methParams.length == params.length) {
              boolean match = true;
              for (int j = 0; j < params.length; j++) {
                match = match && params[j].equals(methParams[j]);
              }
              if (match) nameList.add(signame);
            }
          } else {
            nameList.add(signame);
          }
        }
        else if ((!exact) && name.toUpperCase().startsWith(memberNameUC)) {
          nameList.add(signame);
        } else {
          /* Not a Match */
        }
      }
    } catch(IllegalArgumentException e) {
      Shell.warning("Can not load an array of " + objArray.getClass()); // stack trace not needed
    }

    String sortedList[] = nameList.toArray(new String[0]);
    java.util.Arrays.sort(sortedList); // Sort the list of names

    return sortedList;
  }

  /** Gets the class for a given name. <br>
      If given the name of a primitive the class representing that primitive will be
      returned. If given a name that is not fully resolved (like "String") it will try to
      resolve it (to "java.lang.String") from a set of known packages.
      @param str The name of the class.
      @return The class or null if not found.
   */
  private static Class<?> getClass (String str) {
    final String[] PACKAGE_SEARCH = { "", /* No Package */
                                      "java.lang.", "java.util.",
                                      "nxm.sys.inc.", "nxm.sys.lib.",
                                      "nxm.sys.libg.", "nxm.sys.libm.",
                                      "nxm.sys.net." };
    Class<?> clazz = null;

    /* Convert primitives to their wrapper classes. */
    if (str.equals("long"   )) clazz = java.lang.Long.TYPE;
    if (str.equals("int"    )) clazz = java.lang.Integer.TYPE;
    if (str.equals("short"  )) clazz = java.lang.Short.TYPE;
    if (str.equals("char"   )) clazz = java.lang.Character.TYPE;
    if (str.equals("byte"   )) clazz = java.lang.Byte.TYPE;
    if (str.equals("double" )) clazz = java.lang.Double.TYPE;
    if (str.equals("float"  )) clazz = java.lang.Float.TYPE;
    if (str.equals("boolean")) clazz = java.lang.Boolean.TYPE;


    /* Get the class based on the name, prefix the known packages if
       not initially found.
     */
    for (int i = 0; (clazz == null) && (i < PACKAGE_SEARCH.length); i++) {
      try {
        clazz = Class.forName(PACKAGE_SEARCH[i] + str);

      } catch(Exception e) {
        /* Ignore */
      }
    }

    return clazz;
  }
}
